module getstartendexp_mod
!
! Changes
!
! 02/10/2000 - (AMK) - Implemented a call to OAL_getIPPVstring in order to get the filter form the summary file.
!              This information is now stored in the summary file - previously is wasn't. The filter information 
!              was retrieved from the PEH file directly and decoded in the correct filter string. THe only danger
!              with this method was that it would fail if the filter definitions were to change. This new method is
!              thus more secure.
!              A call to OAL_getIPPVstring is made to return the filter as a string. This value sets the new subroutine
!              parameter - filterString. OMPREP and OMTHCONV can now access this string directly.
! 10/10/2000 - (AMK) - Changed the typeP,A513 and A514 definitions so that task condforms to the ICD.
!              Currently data is being issued that conforms with the ICD and has 32 bit inofrmation in the
!              TYPE column - before 16 bit information was incorrectly being stored. The code had previously been 
!              fixed so that the task would work. Our change will temporarily cause the ODF's upto cycle 149 (we think)
!              apart from the reprocessed cycles to fail. WE will change the code ASAP to use the SID column instead. 
!              All ODF's should then work!!!!!!
! 
! 29/02/2001 - (CB) Code added to prevent program from crashing if the start or end times are outside the range of values
!              in the peh file. Warning messages added (fixes SPR 2249) 
! 06/03/2001 - Old routime getstartendexp renamed test and the new version uses oal calls to get start and end times.
! 03/05/2001 - Added argument "instrumentMode" to get the omfchain working- previously fixed to "IMAGING"
!-
!
!------------------------------------------------------------------------------------------------------------------
  
  use dal
  use oal
  use errorHandling
!  use sla
  implicit none

  private

  public getstartendexp
!  public test  
contains

 
!********************************************************
! subroutine getstartendexp - created 05-03-2001
! Replaces the original version above to fix a bug.
!********************************************************
subroutine getstartendexp(nphFile, exposureNumber, instrumentMode, &
                         OBTstart, OBTend, expStart, &
                         expEnd, filterString)
 
    character(len=255), intent(in)             :: nphFile             ! OM ODF Non-Periodic Housekeeping filename.
    character(len=255)                         :: fileName            ! Input file filename.
    integer, intent(in)                        :: instrumentMode      ! Instrument mode
    character(len=*), parameter                :: odfDir = '../odf'
    character(len=*),intent(out)               :: expStart            ! True start time
    character(len=*),intent(out)               :: expEnd              ! True end time
    character(len=*),intent(out)               :: filterString        ! Exposure filter

!---Test VNY 28.06.2001
    character(len=255)                           :: testdataString
   


    type(ExposureInfoType)                     :: exposureInfo        ! Structure to hold the exposure information.
    integer,intent(in)                         :: exposureNumber      ! The exposure number we are searching for.
    integer                                    :: nCounter            ! Counter
    integer                                    :: nExps               ! For OAL calls
    integer                                    :: nCCDs               ! For OAL calls
    integer                                    :: nodeNr              ! For OAL calls

    real(kind=double),intent(out)              :: obtStart            ! On Board Time start time.
    real(kind=double),intent(out)              :: obtEnd              ! On Board Time end time.
 
    

    !---------------------------------------------------------------------------
    call OAL_setState(OM, instrumentMode)

    !*****************************************
    ! Get number of exposures in the ODF
    !*****************************************
    nExps = OAL_selectScope(OM, instrumentMode)  ! Changed IMAGING to FAST


    !*******************************************************************
    ! Now loop through all the exposures until the required one is found
    !*******************************************************************    
    
    do nCounter = 1, nExps

       !*************************************
       ! select data from exposure nCounter
       !*************************************
       nCCDs = OAL_selectExposure(nCounter);

       !*************************************
       !   select data from central CCD
       !*************************************       
       call OAL_selectFile(0, fileName)
       call OAL_getState(nodeNr = nodeNr)

       !************************************************
       !   inquire information about exposure nCounter
       !
       ! need this
       !************************************************

       nCCds = OAL_selectExposure(nCounter)

       !******************
       ! Is it this one?
       !******************
       call OAL_setState(nodeNr = 0)
       call OAL_exposureInfo(exposureInfo)

!write(*, *) "exposureInfo%number=", exposureInfo%number, " exposureNumber=",  exposureNumber
       if(exposureInfo%number .eq. exposureNumber) then

          
          !****************************************************************************
          !
          ! AMK - 02/10/2000
          !
          ! Make a call to the OAL to retrieve the filter string from the summary File.
          !
          !*****************************************************************************
          filterString  = OAL_getIPPVstring("FILTER")

          !filterNumber = OAL_getIPPVintf("DATA_MODE_2")
          testdataString = OAL_getIPPVstring("DATA_MODE_1")

          !**************************************************************************************
          !
          ! Set the start and end time of the exposure as obtained from the exposure information.
          !***************************************************************************************
          OBTstart = ToTimeTag(exposureInfo%actualStartTime)
          OBTend  = ToTimeTag(exposureInfo%actualEndTime)
          expStart = exposureInfo%actualStartTime
          expEnd = exposureInfo%actualEndTime
          return
       end if


    end do

    !*************************************************************************
    ! Error if reached here- Cound not find the relevant exposure information
    !*************************************************************************
    call fatal("getstartendexp_mod - could not obtain exposure information", &
               "No exposureInfo%number equals the exposure number")
  
return
write(*, *) "nphfile=", nphFile


end subroutine getstartendexp

end module getstartendexp_mod























