!****************************************************************************************
! Given an input image, rotates it by the given angle
! and then save it to a fits file
!
!*****************************************************************************************  
module omdistortion_mod

  use oal
  use cal
 
  use sasdal
  use caloaldefs

  use errorHandling
  use sla
  use omconstants_mod
implicit none


public
   

   real(single), pointer :: rawx(:) => null(), rawy(:)=>null()
 
   real(single), pointer :: deltax(:) => null(), deltay(:)=>null()    
   real(single), pointer :: sigmax(:) => null(), sigmay(:)=>null()    

   real(single) :: x0, y0
   real(double) :: dxBinning, dyBinning

contains

subroutine InitialiseDistortion(x00, y00, xBinning, yBinning)

    integer, intent(in) :: x00, y00
    integer(int32), intent(in) :: xBinning, yBinning



    x0 = dble(x00)
    y0 = dble(y00)
    dxBinning = dble(xBinning)
    dyBinning = dble(yBinning)

    
    if (.not. associated(rawx)) allocate(rawx(0 : 0))
    if (.not. associated(rawy)) allocate(rawy(0 : 0))
    if (.not. associated(deltax)) allocate(deltax(0 : 0))
    if (.not. associated(deltay)) allocate(deltay(0 : 0))
    if (.not. associated(sigmax)) allocate(sigmax(0 : 0))
    if (.not. associated(sigmay)) allocate(sigmay(0 : 0))


end subroutine InitialiseDistortion


!**************************************************************
! Corrects position x and y for distortion
!**************************************************************
subroutine RemoveDistortion(x, y, xOut, yOut)

    real(double), intent(in) :: x, y
    real(double), intent(out) :: xOut, yOut
    integer(int32) :: mode
   

    !****************************************************! 
    ! compute absolute position of object within OM frame
    !****************************************************
    rawx(0) = x * dxBinning + x0   
    rawy(0) = y * dyBinning + y0 

    !***************************************************
    ! correct positions for pincushion/barrel distortion
    !***************************************************




    mode = 1
    call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, mode)


    xOut = x - dble(deltax(0)) / dxBinning ! - sign removes distortion    
    yOut = y - dble(deltay(0)) / dyBinning ! - sign removes distortion

            
end subroutine RemoveDistortion



!**************************************************************
! Corrects position x and y for distortion
!**************************************************************
subroutine RemoveDistortion1(x, y, xOut, yOut)

    real(single), intent(in) :: x, y
    real(double), intent(out) :: xOut, yOut
    integer(int32) :: mode 
  
 

    !****************************************************! 
    ! compute absolute position of object within OM frame
    !****************************************************
    rawx(0) = x   
    rawy(0) = y 
   
    !***************************************************
    ! correct positions for pincushion/barrel distortion
    !***************************************************
  
    mode = 1
    call CAL_omDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, mode)


 
    xOut = x - real(deltax(0)) - x0 ! - sign removes distortion    
    yOut = y - real(deltay(0)) - y0 ! - sign removes distortion


end subroutine RemoveDistortion1



!**************************************************************
! Adds distortion
!**************************************************************
subroutine AddDistortion(x, y, xOut, yOut)

    real(double), intent(in) :: x, y
    real(double), intent(out) :: xOut, yOut
    integer(int32) :: mode

  
    !****************************************************! 
    ! compute absolute position of object within OM frame
    !****************************************************
    rawx(0) = x * dxBinning + x0   
    rawy(0) = y * dyBinning + y0 

    !***************************************************
    ! correct positions for pincushion/barrel distortion
    !***************************************************
   

    mode = 1
    call CAL_omInverseDistortion(rawx, rawy, deltax, deltay, sigmax, sigmay, mode)


    xOut = x + dble(deltax(0)) / dxBinning     
    yOut = y + dble(deltay(0)) / dyBinning 

end subroutine AddDistortion


end module omdistortion_mod





























































































































































































































































