module ommode_mod

use omnumerics_mod

!****************************************************
!
!
!***************************************************
public




interface omMode
    module procedure omMode1, omMode2, ommode3
end interface omMode


contains





!********************************************************
!
!*******************************************************
real(single) function ommode1(image, nx, ny)
    real(single), pointer :: image(:, :)
    integer(int16), intent(in) :: nx, ny

    integer(int16) :: i, j
    integer :: n
    real(single), pointer :: data(:) => null()


    data => allocateSingle(1, (nx + 1) * (ny + 1), "data")

    n = 0
    do j = 0, ny - 1

        do i = 0, nx - 1
            if (image(i, j) .ge. 0.0) then 
                n = n + 1
                data(n) = image(i, j)
            end if
        end do
    end do

    ommode1 = ommode2(n, data)



    deallocate(data)

end function ommode1



!********************************************************
!
!*******************************************************
real(single) function ommode3(ndata, data, nsims)

    integer, intent(in) :: ndata
    real(single) :: data(:)
    integer, intent(in) :: nsims
    integer :: i, j, k     
    real(single), pointer :: tempdata(:) => null()

    real(single), pointer :: modedata(:) => null()
    real(single) :: number

    
    tempdata => allocateSingle(1, ndata, "tempData")
    modedata => allocateSingle(1, nsims, "modeData")

    do i = 1, nsims

        do j = 1, ndata
            call random_number( number )
           
            k = int(number * real(ndata - 1) + 0.0001) + 1
        
            tempData(j) = data(k)
        end do

        modeData(i) = ommode2(ndata, tempdata)

    end do
    !********************************************************
    ! Find the mean
    !********************************************************
    ommode3 = sum(dble(modeData)) / dble(nsims)


    deallocate(tempdata, modedata)

end function ommode3
!********************************************************
!
!*******************************************************
real(single) function ommode2(ndata, data)

    integer, intent(in) :: ndata
    real(single) :: data(:)
  
    integer i, iStart, iFinish, size, minSize, n 
    real(single) :: constant, mean, median     
    real(single), pointer :: tempdata(:) => null()
     
    n = count( data(1 : ndata) >= 0.0)
    constant = 0.5
    tempdata => allocateSingle(1, n, "tempData")


    n = 0
    do i = 1, nData
        if (data(i) >= 0.0) then
            n = n + 1
            tempdata(n) = data(i)
        end if
    end do

    call HeapSort( tempData, n)
    
    mean = sum(dble(tempData)) / dble(n)
    median = SortedMedian(tempData, n)


    minSize = max(5, n / 20)
    
    iStart = 1
    iFinish = n

    do i = 1, 1000
   
 
        call maxInterval(tempdata, iStart, iFinish, constant)
        size = iFinish - iStart + 1
        
       	

	if (size <= minSize .or. size == n) then   
        
            
	    ommode2 = sum(tempdata(iStart : ifinish) ) / dble(size)

!            write(*, *) "image mode=", ommode2, " no of iterations=", i
!            write(*, *) "mean=", mean, " median=", median
            exit

        end if
        
        n = size

    end do

    deallocate(tempdata)


end function ommode2	
!*********************************************************************************
!
!********************************************************************************* 
subroutine maxInterval(data, iStart, ifinish,  constant)
    real(single), pointer :: data(:)
    integer, intent(inout) :: iStart, iFinish
    real(single), intent(in) :: constant
  
    integer :: i, j, maxCount, noBins, iu, count1, n, n1 
    real(single) :: width, halfWidth, wMin 
    real(single) :: fPLus, d1, d2

    integer :: istart1, ifinish1, iu1
    
    istart1 = istart
    ifinish1 = ifinish

 
    width = data(iFinish) - data(iStart)

    if (width < 0.1) return

    halfWidth = width * constant


    wMin = 10000000000.0    

    maxCount = 0
    noBins = 0

    n = iFinish - iStart + 1 ! The number of points
     
   
    iu = iStart
    do i = iStart, iFinish - 1
   
    
       
        fPlus  = data(i) + halfWidth
        iu1 = iu
        iu = iFinish
        do j = iu1, iFinish
            if(data(j) > fPlus) then

                iu = j - 1
                exit
            end if
             
        end do
         
    	
	   
        count1 = iu - i + 1
      	 
	if (count1 == n) then ! Finished
            istart = i
            iFinish = iu
	    return
        end if

    
        if (count1 > maxCount) then ! Maximum no of points 
	
	 
	       
            wMin = data(iu) - data(i) ! The new minimum interval width
                   
            noBins = 1
	    iStart1 = i
            iFinish1 = iu
 
            maxCount = count1


        elseif (count1 == maxCount) then
	


	    width = data(iu) - data(i)
            if (width < wMin)then
	   
	        wMin = width   ! Smallest bin width  

                 
              
                noBins = 1
                
                iStart1 = i
                iFinish1 = iu
 
                  
            else if(abs( width - wMin) < 0.0000001) then ! Bin widths equals minimum bin width

                noBins = noBins + 1
                iFinish1 = iu
 

	    end if
      
        end if 

        if(iu == iFinish) exit
    end do 

    if (noBins == 1) then ! Only 1 bin
  

        iStart = iStart1
        iFinish = iFinish1

        return

    end if


    !********************************************************************
    ! More than one bin with the same width and count
    !********************************************************************
    n1 = iFinish1 - iStart1 + 1    

    

    iStart = istart1
    iFinish = ifinish1

    if (n == n1) then

        d1 = data(istart + 1) - data(istart)
        d2 = data(ifinish) - data(ifinish - 1) 
        if ( d1 < d2 ) then
            iFinish = iFinish - 1
        elseif(d1 > d2) then
            istart = istart + 1
        else
            istart = istart + 1
            ifinish = ifinish - 1
        end if  
    end if  


end subroutine Maxinterval

end module ommode_mod



































