!******************************************************************
! Module omdetectnumerics_mod
!
! Contains various subroutines for numerical procedures.
!******************************************************************
module omnumerics_mod
  use errorHandling
  use OmAllocation_mod
  use omconstants_mod
!  use sort_mod


interface HeapSort
    module procedure HeapSort1, HeapSort2, HeapSort3
end interface HeapSort
   
interface Locate
    module procedure Locate1, Locate2
end interface Locate
   


interface IndexSort
    module procedure IndexSort1, IndexSort2, IndexSort3
end interface IndexSort
   
interface ComputeAverage
    module procedure ComputeAverage1, ComputeAverage2, ComputeAverage3, ComputeAverage4
end interface ComputeAverage

interface ComputeStatistics
    module procedure ComputeStatistics1, ComputeStatistics2
end interface  ComputeStatistics

interface RobustMean
    module procedure RobustMean1, RobustMean2
end interface RobustMean

interface a15
    module procedure a151, a152
end interface a15

interface h15
    module procedure h151, h152
end interface h15

interface smad
    module procedure smad1, smad2
end interface smad

interface medianValue
    module procedure medianValue1, medianValue2
end interface medianValue



contains


!********************************
! function Angle
! 
!********************************
real(single) function Angle(x, y, xy)
    real (single), intent(in) :: x, y, xy
    real (single) :: theta1, theta2 

    if(x .ne. y) then
        theta1 = atan(2 * xy / (x - y)) / 2.0
        theta2 = theta1 + HALF_PI
        if(theta2 .gt. HALF_PI) theta2 = theta2 - PI
        if(xy .ge. 0.0) then
            if(theta1 .ge. 0.0) then
                Angle = theta1
            else
                Angle = theta2
            end if
        else
            if(theta1 .lt. 0.0) then
                Angle = theta1
            else
                Angle = theta2
            end if
        end if
    else
        Angle = sign(real(HALF_PI), xy) / 2.0
    end if
end function Angle


!***************************************************************
! Function MedianValue
! Find the median of X(1), ... , X(N), 
! using as much of the quicksort
! algorithm as is needed to isolate it.
!**********************************************************************
real(single) Function MedianValue1(data, n)

    IMPLICIT NONE
    INTEGER(int32), INTENT(IN)                :: n

  
    REAL(single) :: data(:)
  
  
    !************************************
    ! Local variables
    !***********************************

    REAL(single)    :: temp, xhi, xlo, xmax, xmin
    LOGICAL (bool) :: odd
    INTEGER (int32) :: hi, lo, nby2, nby2p1, mid, i, j, k
    real(single), dimension(:), pointer :: x => null()
  
!write(*, *) "medianvalue1 line 115 n=", n
  



    if( minval(data(1 : n)) .eq. maxval(data(1 : n))) then
        medianvalue1 = data(1)
        return

    end if




    nby2 = ishft(n, -1)
    nby2p1 = nby2 + 1
   
    !******************************************************
    ! HI & LO are position limits encompassing the median.
    !******************************************************
    IF (n .eq. 2 * nby2) then 
        odd = .false.
    else
        odd = .false.
    end if
    lo = 1
    hi = n
    IF (n .lt. 3) THEN
        IF (n .lt. 1) THEN
            MedianValue1 = 0.0_double
            RETURN
        END IF
        MedianValue1 = data(1)
        IF (n .eq. 1) RETURN
        MedianValue1 = 0.5 * (MedianValue1 + data(2))
        RETURN
    END IF

  
    !**********************************************
    ! Copy the data array to another one so that 
    ! original one is unchanged
    !**********************************************

    x => AllocateSingle(1, n, "x (MedianValue1)")
    x = data(1 : n)

!write(*, *) "medianvalue1 line 147 n=", n

    !********************************************
    ! Find median of 1st, middle & last values
    !********************************************
10 mid = (lo + hi) / 2
    MedianValue1 = x(mid)
    xlo = x(lo)
    xhi = x(hi)

    IF (xhi .lt. xlo) THEN          ! Swap xhi & xlo
        temp = xhi
        xhi = xlo
        xlo = temp
    END IF
    IF (MedianValue1 .gt. xhi) THEN
        MedianValue1 = xhi
    ELSE IF (MedianValue1 < xlo) THEN
        MedianValue1 = xlo
    END IF

    !****************************************************
    ! The basic quicksort algorithm to move all values 
    ! <= the sort key (XMED) to the left-hand end, and 
    ! all higher values to the other end.
    !****************************************************
    i = lo
    j = hi

50  DO

!write(*, *) "Line 193 medianvalue1 lo=", lo, " hi=", hi, " n=", n, " i=", i, " j=", j
        
        IF (x(i) .ge. MedianValue1) EXIT
        i = i + 1
    END DO
    DO
        IF (x(j) .le. MedianValue1) EXIT
        j = j - 1
    END DO

    IF (i .lt. j) THEN
        temp = x(i)
        x(i) = x(j)
        x(j) = temp
        i = i + 1
        j = j - 1

        !***************************************
        ! Decide which half the median is in.
        !***************************************

        IF (i .le. j) goto 50
    END IF

!write(*, *) "Line 217 medianvalue1 lo=", lo, " hi=", hi, " n=", n, " i=", " j=", j



    IF (.NOT. odd) THEN
        IF (j .eq. nby2 .AND. i .eq. nby2p1) GO TO 130
        IF (j .lt. nby2) lo = i
        IF (i .gt. nby2p1) hi = j
        IF (i .ne. j) GO TO 100
        IF (i .eq. nby2) lo = nby2
        IF (j .eq. nby2p1) hi = nby2p1
    ELSE
        IF (j .lt. nby2p1) lo = i
        IF (i .gt. nby2p1) hi = j
        IF (i .ne. j) GO TO 100
        !****************************************
        ! Test whether median has been isolated.
        !****************************************
        IF (i .eq. nby2p1) then
            deallocate(x)
            RETURN
        end if
    END IF

!write(*, *) "Line 241 medianvalue1 lo=", lo, " hi=", hi, " n=", n, " i=", " j=", j

100 IF (lo .lt. hi - 1) GO TO 10

    IF (.NOT. odd) THEN
        MedianValue1 = 0.5 * (x(nby2) + x(nby2p1))
        deallocate(x)
        RETURN
    END IF

!write(*, *) "Line 251 medianvalue1 lo=", lo, " hi=", hi, " n=", n, " i=", " j=", j

    temp = x(lo)
    IF (temp .gt. x(hi)) THEN
        x(lo) = x(hi)
        x(hi) = temp
    END IF
    MedianValue1 = x(nby2p1)
    deallocate(x)
    RETURN

    !***************************************************************
    ! Special case, N even, J = N/2 & I = J + 1, so the median is
    ! between the two halves of the series.   Find max. of the first
    ! half & min. of the second half, then average.
    !***************************************************************
    130 xmax = x(1)
    DO k = lo, j
        xmax = max(xmax, x(k))
    END DO
    xmin = x(n)

    DO k = i, hi
        xmin = min(xmin, x(k))
    END DO

    MedianValue1 = 0.5 * (xmin + xmax)
    deallocate(x)

end function MedianValue1


!***************************************************************
! Function MedianValue
! Find the median of X(1), ... , X(N), 
! using as much of the quicksort
! algorithm as is needed to isolate it.
!**********************************************************************
real(double) Function MedianValue2(data, n)

    IMPLICIT NONE
    INTEGER(int32), INTENT(IN)                :: n
     REAL(double) :: data(:)
  
    
    !************************************
    ! Local variables
    !***********************************

    REAL(double)    :: temp, xhi, xlo, xmax, xmin
    LOGICAL (bool) :: odd
    INTEGER (int32) :: hi, lo, nby2, nby2p1, mid, i, j, k







    if( minval(data(1 : n)) .eq. maxval(data(1 : n))) then
        medianvalue2 = data(1)
        return

    end if





   
    nby2 = ishft(n, -1)
    nby2p1 = nby2 + 1
    odd  = .true.

    !******************************************************
    ! HI & LO are position limits encompassing the median.
    !******************************************************
    IF (n .eq. 2 * nby2) odd = .false.
    lo = 1
    hi = n
    IF (n .lt. 3) THEN
        IF (n .lt. 1) THEN
            MedianValue2 = 0.0_double
            RETURN
        END IF
        MedianValue2 = data(1)
        IF (n .eq. 1) RETURN
        MedianValue2 = 0.5 * (MedianValue2 + data(2))
        RETURN
    END IF









    !**********************************************
    ! Copy the data array to another one so that 
    ! original one is unchanged
    !**********************************************


    !********************************************
    ! Find median of 1st, middle & last values
    !********************************************
10 mid = (lo + hi)/2
    MedianValue2 = data(mid)
    xlo = data(lo)
    xhi = data(hi)
    IF (xhi .lt. xlo) THEN          ! Swap xhi & xlo
        temp = xhi
        xhi = xlo
        xlo = temp
    END IF
    IF (MedianValue2 .gt. xhi) THEN
        MedianValue2 = xhi
    ELSE IF (MedianValue2 < xlo) THEN
        MedianValue2 = xlo
    END IF
    !****************************************************
    ! The basic quicksort algorithm to move all values 
    ! <= the sort key (XMED) to the left-hand end, and 
    ! all higher values to the other end.
    !****************************************************
    i = lo
    j = hi
50  DO
        IF (data(i) .ge. MedianValue2) EXIT
        i = i + 1
    END DO
    DO
        IF (data(j) .le. MedianValue2) EXIT
        j = j - 1
    END DO
    IF (i .lt. j) THEN
        temp = data(i)
        data(i) = data(j)
        data(j) = temp
        i = i + 1
        j = j - 1

        !***************************************
        ! Decide which half the median is in.
        !***************************************

        IF (i .le. j) goto 50
    END IF

    IF (.NOT. odd) THEN
        IF (j .eq. nby2 .AND. i .eq. nby2p1) GO TO 130
        IF (j .lt. nby2) lo = i
        IF (i .gt. nby2p1) hi = j
        IF (i .ne. j) GO TO 100
        IF (i .eq. nby2) lo = nby2
        IF (j .eq. nby2p1) hi = nby2p1
    ELSE
        IF (j .lt. nby2p1) lo = i
        IF (i .gt. nby2p1) hi = j
        IF (i .ne. j) GO TO 100
        !****************************************
        ! Test whether median has been isolated.
        !****************************************
        IF (i .eq. nby2p1) then          
            RETURN
        end if
    END IF
100 IF (lo .lt. hi - 1) GO TO 10

    IF (.NOT. odd) THEN
        MedianValue2 = 0.5 * (data(nby2) + data(nby2p1))
        RETURN
    END IF
    temp = data(lo)
    IF (temp .gt. data(hi)) THEN
        data(lo) = data(hi)
        data(hi) = temp
    END IF
    MedianValue2 = data(nby2p1)
   
    RETURN

    !***************************************************************
    ! Special case, N even, J = N/2 & I = J + 1, so the median is
    ! between the two halves of the series.   Find max. of the first
    ! half & min. of the second half, then average.
    !***************************************************************
    130 xmax = data(1)
    DO k = lo, j
        xmax = max(xmax, data(k))
    END DO
    xmin = data(n)

    DO k = i, hi
        xmin = min(xmin, data(k))
    END DO

    MedianValue2 = 0.5 * (xmin + xmax)
   
end function MedianValue2









!**************************************
! Subroutine HeapSort
!**************************************
subroutine HeapSort1(array, size)
    integer(int32), intent(in) :: size
    real(single) :: array(:)
   

    integer :: i, ir, j, l
    real(single) :: rra
  
    if (size .lt. 2) return


    l = ishft(size, -1) + 1
    ir = size   

    do 
        if (l .gt. 1) then
            l = l - 1
            rra = array(l)   
              
        else
            rra = array(ir)        
            array(ir) = array(1)
          
           
            ir = ir - 1
            
            if (ir .le. 1) then
             
                array(1) = rra
                exit
            end if
        end if

        i = l
        j = l + l
        do while(j .le. ir)
       
            if (j .lt. ir) then
                if(array(j) .lt. array(j + 1)) j = j + 1
            end if
            if (rra .lt. array(j)) then
               
                array(i) = array(j)
                i = j
                j = j + j
            else
                exit
            end if
            
        end do   
          
        array(i) = rra
    end do
  
end subroutine HeapSort1

!**************************************
! Subroutine HeapSort
!**************************************
subroutine HeapSort2(array, size)
    integer(int32), intent(in) :: size
    real(double) :: array(:)
   

    integer :: i, ir, j, l
    real(double) :: rra
  
    if (size .lt. 2) return


    l = ishft(size, -1) + 1
    ir = size   

    do 
        if (l .gt. 1) then
            l = l - 1
            rra = array(l)   
              
        else
            rra = array(ir)        
            array(ir) = array(1)
          
           
            ir = ir - 1
            
            if (ir .le. 1) then
             
                array(1) = rra
                exit
            end if
        end if

        i = l
        j = l + l
        do while(j .le. ir)
       
            if (j .lt. ir) then
                if(array(j) .lt. array(j + 1)) j = j + 1
            end if
            if (rra .lt. array(j)) then
               
                array(i) = array(j)
                i = j
                j = j + j
            else
                exit
            end if
            
        end do   
          
        array(i) = rra
    end do
  
end subroutine HeapSort2

!**************************************
! Subroutine HeapSort
!**************************************
subroutine HeapSort3(array, size)
    integer(int32), intent(in) :: size
    integer(int32) :: array(:)
   

    integer :: i, ir, j, l
    integer(int32) :: rra
  
    if (size .lt. 2) return


    l = ishft(size, -1) + 1
    ir = size   

    do 
        if (l .gt. 1) then
            l = l - 1
            rra = array(l)   
              
        else
            rra = array(ir)        
            array(ir) = array(1)
          
           
            ir = ir - 1
            
            if (ir .le. 1) then
             
                array(1) = rra
                exit
            end if
        end if

        i = l
        j = l + l
        do while(j .le. ir)
       
            if (j .lt. ir) then
                if(array(j) .lt. array(j + 1)) j = j + 1
            end if
            if (rra .lt. array(j)) then
               
                array(i) = array(j)
                i = j
                j = j + j
            else
                exit
            end if
            
        end do   
          
        array(i) = rra
    end do
  
end subroutine HeapSort3





!************************************************************
! Subroutine IndexSort
! Sorts  one-dimensional array(1.. size) int ascending order.
! An index of the sorted order relative to the original
! order is returned.
!************************************************************ 
subroutine IndexSort1(array, index, size)
    integer (int32), intent(in) :: size
    real(single) :: array(:)
    integer (int32)  :: index(:)

    integer (int32) :: i, ir, j, l, irra
    real(single) :: rra
    if (size .eq. 0) return
    if (size < 2) then
        index(1) = 1
    end if

    l = ishft(size, -1) + 1
    ir = size   
    do i = 1, size
        index(i) = i
       
    end do   

    do 
        if (l > 1) then
            l = l - 1
            rra = array(index(l))   
            irra = index(l)    
        else
            rra = array(index(ir))        
            irra = index(ir)
          
            index(ir) = index(1)
            ir = ir - 1
            
            if (ir <= 1) then
             
                index(1) = irra
                exit
            end if
        end if

        i = l
        j = l + l
        do while(j <= ir)
       
            if (j < ir) then
                if(array(index(j)) < array(index(j + 1))) j = j + 1
            end if
            if (rra < array(index(j))) then
               
                index(i) = index(j)
                i = j
                j = j + j
            else
                exit
            end if
            
        end do   
          
        index(i) = irra
    end do
   
end subroutine IndexSort1


!************************************************************
! Subroutine IndexSort
! Sorts  one-dimensional array(1.. size) int ascending order.
! An index of the sorted order relative to the original
! order is returned.
!************************************************************ 
subroutine IndexSort2(array, index, size)
    integer(int32), intent(in) :: size
    real(double) :: array(:)
    integer(int32) :: index(:)

    integer(int32) :: i, ir, j, l, irra
    real(double) :: rra
    if (size .eq. 0) return
    if (size < 2) then
        index(1) = 1
    end if

    l = ishft(size, -1) + 1
    ir = size   
    do i = 1, size
        index(i) = i
       
    end do   

    do 
        if (l > 1) then
            l = l - 1
            rra = array(index(l))   
            irra = index(l)    
        else
            rra = array(index(ir))        
            irra = index(ir)
          
            index(ir) = index(1)
            ir = ir - 1
            
            if (ir <= 1) then
             
                index(1) = irra
                exit
            end if
        end if

        i = l
        j = l + l
        do while(j <= ir)
       
            if (j < ir) then
                if(array(index(j)) < array(index(j + 1))) j = j + 1
            end if
            if (rra < array(index(j))) then
               
                index(i) = index(j)
                i = j
                j = j + j
            else
                exit
            end if
            
        end do   
          
        index(i) = irra
    end do
   
end subroutine IndexSort2

!************************************************************
! Subroutine IndexSort
! Sorts  one-dimensional array(1.. size) int ascending order.
! An index of the sorted order relative to the original
! order is returned.
!************************************************************ 
subroutine IndexSort3(array, index)
    real(single) :: array(:)
    integer (int32) :: index(:)

    integer (int32) :: i, ir, j, l, irra
    real(single) :: rra
 
   integer (int32) :: k
   k = size(array)

    if (k .eq. 0) return
    if (k < 2) then
        index(1) = 1
    end if

    l = ishft(k, -1) + 1
    ir = k   
    do i = 1, k
        index(i) = i
       
    end do   

    do 
        if (l > 1) then
            l = l - 1
            rra = array(index(l))   
            irra = index(l)    
        else
            rra = array(index(ir))        
            irra = index(ir)
          
            index(ir) = index(1)
            ir = ir - 1
            
            if (ir <= 1) then
             
                index(1) = irra
                exit
            end if
        end if

        i = l
        j = l + l
        do while(j <= ir)
       
            if (j < ir) then
                if(array(index(j)) < array(index(j + 1))) j = j + 1
            end if
            if (rra < array(index(j))) then
               
                index(i) = index(j)
                i = j
                j = j + j
            else
                exit
            end if
            
        end do   
          
        index(i) = irra
    end do
   
end subroutine IndexSort3


!***************************
! Function SortedMedian
! Returns the median for
! a sorted set of data
!***************************
real(single) function SortedMedian(array, size)
    integer(int32), intent(in) :: size
    real(single)  :: array(:)
    integer(int32) :: k

    if (size .eq. 0) then 
        SortedMedian = 0.0
        return
    end if
  

    if (size .eq. 1) then 
        SortedMedian = array(1)
        return
    end if

    k = ishft(size, -1)
    if (ishft(k, 1) .eq. size) then   ! Even number
        SortedMedian = (array(k) + array(k + 1)) / 2.0
    else
        SortedMedian = array(k + 1)
    end if 

end function SortedMedian    


!***************************
! Function RobustMean
! Returns the mean for
! a sorted set of data
!***************************

real(single) function RobustMean1(array, size, std)
    integer(int32), intent(in) :: size
    real(single)  :: array(:)
    real(single), intent(out) :: std
   
    RobustMean1 = RobustMean2(array, size, 0.25, std)
    



end function RobustMean1


!***************************
! Function RobustMean
! Returns the mean for
! a sorted set of data
!***************************

real(single) function RobustMean2(array, size, frac, std)
    integer(int32), intent(in) :: size
    real(single), intent(in) :: frac
    real(single)  :: array(:)
    real(single), intent(out) :: std
    integer(int32) :: kl, ku, n
    


    if (n >= 12) then
        kl = int( real(size) * frac + 0.00001) + 1
        ku = int( real(size) * (1.0 - frac) + 0.00001)
        n = ku - kl + 1
    else
        kl = 1
        ku = size
        n = size
    end if

 
    

    call ComputeStatistics(n, array(kl : ku), robustMean2, std)
    
    std = sqrt(std)

end function RobustMean2






!***************************
! Function RobustSortedMedian
! Returns the median for
! a sorted set of data
!***************************
real(single) function RobustMedian(array, size)
    integer(int32), intent(in) :: size
    real(single)  :: array(:)
    integer(int32) :: kl, ku, n
    

    if (size .eq. 0) then 
        RobustMedian = 0.0
        return
    end if
  

    if (size .eq. 1) then 
        RobustMedian = array(1)
        return
    end if

    if (n >= 10) then
        kl = int(real(size * 0.1) + 0.00001)
        ku = int(real(size * 0.9) + 0.00001)
        n = ku - kl + 1
    else
        kl = 1
        ku = size
        n = size
    end if

    robustMedian = MedianValue(array(kl : ku), n)



end function RobustMedian







!***************************
! Function SortedMedian
! Returns the median for
! a sorted set of data
! Ignnore first and last 10% of data
!***************************
real(single) function RobustSortedMedian(array, size)
    integer(int32), intent(in) :: size
    real(single)  :: array(:)
    integer(int32) :: k, k10, size1

    if (size .eq. 0) then 
        RobustSortedMedian = 0.0
        return
    end if
  

    
    if (size .eq. 1) then 
        RobustSortedMedian = array(1)
        return
    end if

    if (size > 9) then
        k10 = int(real(size * 0.1) + 0.00001)

        size1 = size - 2 * k10
    else
        size1 = size
        k10 = 0
    end if
    
    k = ishft(size1, -1)

    if (ishft(k, 1) .eq. size1) then   ! Even number
        k = k + k10
        RobustSortedMedian = (array(k) + array(k + 1)) / 2.0
    else
        RobustSortedMedian = array(k + k10 + 1)
    end if 

end function RobustSortedMedian    


!**********************************
! ComputeAverage
! Calculates the weighted average and
! standard deviation
!**********************************
subroutine ComputeAverage1(data, error, size, mean, standardDeviation)
    integer(int32), intent(in) :: size        
    real(single) :: data(:), error(:)    
    real(single), intent(out) :: mean, standardDeviation
    real(double) :: weight
    real(double) :: sum1, sum2, sum3
    real(double) :: d, difference
    logical :: useWeights
    integer i

    useWeights = .true.
    weight = 1.0_double

    if (size .eq. 1) then
        mean = data(1)
        standardDeviation = error(1)
        return
    end if

    sum1 = 0.0_double
    sum2 = 0.0_double
    sum3 = 0.0_double
 

    do i = 1, size
        if (error(i) .le. 0.0) then
            useWeights = .false.
            sum1 = dble(size)
            sum3 = dble(size)
            exit
        end if
    end do 

    
    do i = 1, size
        if (useWeights) then
            weight = 1.0_double / (error(i) * error(i))
            sum1 = sum1 + weight
            sum2 = sum2 + data(i) * weight
            sum3 = sum3 + weight * weight
        else
            sum2 = sum2 + data(i)
          
        end if

    end do

    mean = sum2 / sum1  ! The mean

    sum2 = 0.0d0 
    do i = 1, size
        difference = data(i) - mean       
        if (useWeights) then
            weight = 1.0_double / (error(i) * error(i))
            sum2 = sum2 + weight * difference * difference
        else
            sum2 = sum2 + difference * difference
        end if 
    end do
    if (useWeights) then
        d = sum1 - sum3 / sum1
    else
        d = dble(size - 1)
    end if
    standardDeviation = dsqrt(sum2 / d) ! The standard deviation

end subroutine ComputeAverage1
!**********************************
! ComputeAverage
! Calculates the weighted average and
! standard deviation
!**********************************
subroutine ComputeAverage2(data, size, mean, standardDeviation)
    integer(int32), intent(in) :: size        
    real(single) :: data(:)    
    real(single), intent(out) :: mean, standardDeviation
 
    real(double) :: sum2
    real(double) :: d, difference
    
    integer i



    if (size .eq. 1) then
        mean = data(1)
        standardDeviation = 0.0
        return
    end if



    mean = sum(data(1 : size)) / dble(size)  ! The mean

    sum2 = 0.0d0 
    do i = 1, size
        difference = data(i) - mean       
      
        sum2 = sum2 + difference * difference
        
    end do
    
    d = dble(size - 1)
    
    standardDeviation = dsqrt(sum2 / d) ! The standard deviation

end subroutine ComputeAverage2




!**********************************
! ComputeAverage
! Calculates the weighted average and
! standard deviation
!**********************************
subroutine ComputeAverage3(data, error, size, mean, standardDeviation)
    integer(int32), intent(in) :: size        
    real(double) :: data(:), error(:)    
    real(double), intent(out) :: mean, standardDeviation
    real(double) :: weight
    real(double) :: sum1, sum2, sum3
    real(double) :: d, difference
    logical :: useWeights
    integer i

    useWeights = .true.
    weight = 1.0_double

    if (size .eq. 1) then
        mean = data(1)
        standardDeviation = error(1)
        return
    end if

    sum1 = 0.0_double
    sum2 = 0.0_double
    sum3 = 0.0_double
 

    do i = 1, size
        if (error(i) .le. 0.0) then
            useWeights = .false.
            sum1 = dble(size)
            sum3 = dble(size)
            exit
        end if
    end do 

    
    do i = 1, size
        if (useWeights) then
            weight = 1.0_double / (error(i) * error(i))
            sum1 = sum1 + weight
            sum2 = sum2 + data(i) * weight
            sum3 = sum3 + weight * weight
        else
            sum2 = sum2 + data(i)
          
        end if

    end do

    mean = sum2 / sum1  ! The mean

    sum2 = 0.0d0 
    do i = 1, size
        difference = data(i) - mean       
        if (useWeights) then
            weight = 1.0_double / (error(i) * error(i))
            sum2 = sum2 + weight * difference * difference
        else
            sum2 = sum2 + difference * difference
        end if 
    end do
    if (useWeights) then
        d = sum1 - sum3 / sum1
    else
        d = dble(size - 1)
    end if
    standardDeviation = dsqrt(sum2 / d) ! The standard deviation

end subroutine ComputeAverage3
!**********************************
! ComputeAverage
! Calculates the weighted average and
! standard deviation
!**********************************
subroutine ComputeAverage4(data, size, mean, standardDeviation)
    integer(int32), intent(in) :: size        
    real(double)  :: data(:)    
    real(double), intent(out) :: mean, standardDeviation
  
    real(double) :: sum2
    real(double) :: d, difference
  
    integer i


    if (size .eq. 1) then
        mean = data(1)
        standardDeviation = 0.0
        return
    end if

   
    mean = sum(data) / dble(size)  ! The mean

    sum2 = 0.0d0 
    do i = 1, size
        difference = data(i) - mean       
      
        sum2 = sum2 + difference * difference
        
    end do
    
    d = dble(size - 1)
    
    standardDeviation = dsqrt(sum2 / d) ! The standard deviation

end subroutine ComputeAverage4


!**************************************************
!
!**************************************************
subroutine ComputeStatistics1(n, x, mean, sigma)
    integer, intent(in) :: n
    real(single) :: x(:)
    real(single), intent(out) :: mean, sigma
 

    mean = sum(x(1 : n)) / real(n)   
    sigma = sum((x(1 : n) - mean) ** 2) / real(n - 1)
  
end subroutine ComputeStatistics1

!**************************************************
!
!**************************************************
subroutine ComputeStatistics2(n, x, mean, sigma)
    integer, intent(in) :: n
    real(double) :: x(:)
    real(double), intent(out) :: mean, sigma
   

    mean = sum(x(1 : n)) / real(n)   
    sigma = sum((x(1 : n) - mean) ** 2) / dble(n - 1)





  
end subroutine ComputeStatistics2

!**************************************************
!
!**************************************************
subroutine ComputeStatistics22(n, x, mean, sigma, itermax)
    integer, intent(inout) :: n
    integer, intent(in) :: itermax
    real(single) :: x(:)
    real(single), intent(out) :: mean, sigma
    real(double) :: mean1, sigma1, sigma2

    integer :: i, j, k, m
!    integer, save :: count = 0

    m = n
    do i = 1, iterMax
       
        mean1 = sum(x(1 : m)) / real(m)   
        sigma1 = sum((x(1 : m) - mean) ** 2) / dble(m - 1)



!if (count .lt. 100) then
!count = count + 1
!write(*, *) "Line 1125 iteration ", i, " mean=", mean1, " sigma=", sigma1
!end if        



        k = 0
        sigma2 = 2.0 * sigma1
        do j = 1, m
            if (abs(mean1 - x(j)) .lt. sigma2) then
                k = k + 1
                x(k) = x(j)
            end if
        end do
         
        if (k .eq. m) then
            mean = mean1
            sigma = sigma1 * sigma1
                 
            return
        end if  
  
        if (k .lt. ishft(n, -1)) then
            if (i .eq. 1) then
                mean = mean1
                sigma = sigma1 * sigma1
                 
                return
            else
                                 
                return

            end if
        end if
  
        m = k    
        mean = mean1
        sigma = sigma1 * sigma1
    end do
        
   


    n = m





    
end subroutine ComputeStatistics22




!**************************************************
!
!**************************************************
subroutine ComputeStatistics3(n, x, mean, sigma)
    integer, intent(inout) :: n
    

    real(single) :: x(:)
  

    real(single), intent(out) :: mean, sigma
  
!write(*, *) "Compute statistisc3 line 1342 n=", n, " size x=", size(x)
!write(*, *) "x=", x(1 : n)  
   mean = medianValue(x,  n)
   sigma = sum(abs(x(1 : n) - mean)) / dble(n - 1)
    sigma = sigma * sigma

!write(*, *) "Compute statistisc3 line 1347 n=", n, " size x=", size(x)


end subroutine computestatistics3


!**************************************************
!
!**************************************************
subroutine ComputeStatistics33(n, x, mean, sigma, iterMax)
    integer, intent(inout) :: n
    

    real(single) :: x(:)
    integer :: i, j, k, m   

    real(single), intent(out) :: mean, sigma
    integer, intent(in) :: iterMax
 
    real(single) :: sigma1, sigma2, mean1
integer, save :: count = 0
  
!   mean = medianValue(x(1 : n), n)
!        sigma = sum(abs(x(1 : n) - mean)) / dble(n - 1)
!    sigma = sigma * sigma


!return

    m = n

    


    do i = 1, iterMax
        mean1 = sortedmedian(x(1 : m), m)
        sigma1 = sum(abs(x(1 : m) - mean)) / dble(m - 1)

if (count .lt. 100) then
count = count + 1
write(*, *) "Line 1157 iteration ", i, " mean=", mean1, " sigma=", sigma1
end if        

        k = 0
        sigma2 = 2.0 * sigma1
        do j = 1, m
            if (abs(mean1 - x(j)) .lt. sigma2) then
                k = k + 1
                x(k) = x(j)
            end if
        end do
         
        if (k .eq. m) then
            mean = mean1
            sigma = sigma1 * sigma1
                 
            return
        end if  
  
        if (k .lt. ishft(n, -1)) then
            if (i .eq. 1) then
                mean = mean1
                sigma = sigma1 * sigma1
                 
                return
            else
                                 
                return

            end if
        end if
  
        m = k    
        mean = mean1
        sigma = sigma1 * sigma1
    end do
        
   


    n = m

end subroutine ComputeStatistics33
!********************************************************
!
!*******************************************************
subroutine ComputeStatistics4(n, x, mean, sigma)
    integer, intent(in) :: n
    real(single) :: x(:)
    real(single), intent(out) :: mean, sigma

    real(double) :: weight(n)
    real(double) :: sum1, sum2, sum3
    real(double) :: d, difference
   
    integer(int32) ::  i

    


    sum1 = 0.0_double
    sum2 = 0.0_double
    sum3 = 0.0_double
 
    
    do i = 1, n
        if (x(i) .gt. 0.0) then 
            weight(i) = 1.0_double / dble(x(i))
            sum1 = sum1 + weight(i)
            sum2 = sum2 + dble(x(i)) * weight(i)
            sum3 = sum3 + weight(i) * weight(i)
        else
            weight(i) = 0.0_double
        end if
    end do

    mean = sum2 / sum1  ! The mean

    sum2 = 0.0d0 
    do i = 1, n
        difference = dble(x(i)) - mean       
        sum2 = sum2 + weight(i) * difference * difference
      
    end do
    
    d = sum1 - sum3 / sum1
    
    sigma = sum2 / d ! The variance


end subroutine computestatistics4

!********************************************
! Function Locate1
! Given an array table(1 .. n), and
! given a value x, returns the index such that
! x is between table(j) and table(j+1)
! The search starts at the position jStart
!********************************************
integer function Locate1(size, table, searchValue, jStart)
    implicit none
    integer, intent(in) :: size
    real(single), dimension(:) :: table
    real(single), intent(in) :: searchValue
    integer, intent(inout) :: jStart
    integer :: k, klo, khi

    !********************************************
    ! Check to see if x is less than the minimum
    ! value in the table
    !********************************************
    if (searchValue .lt. table(1)) then
        Locate1 = 1
        return
    end if

    !***************************************
    ! Check to see if x is greater than the
    ! maximum value in the table
    !***************************************
    if (searchValue .gt. table(size)) then
        Locate1 = size
        return
    end if

    klo = jStart
    khi = size

    do while((khi - klo) .gt. 1)
        k = (khi + klo) / 2
        if (table(k) .gt. searchValue) then
            khi = k
        else
            klo = k
        end if
    end do

    !*********************************
    ! klo and khi now bracket the
    ! searchvalue
    !*********************************
    Locate1 = klo
    jStart = klo
end function Locate1
!********************************************
! Function Locate1
! Given an array table(1 .. n), and
! given a value x, returns the index such that
! x is between table(j) and table(j+1)
! The search starts at the position jStart
!********************************************
integer function Locate2(size, table, searchValue, jStart)
    implicit none
    integer, intent(in) :: size
    real(double), dimension(:) :: table
    real(double), intent(in) :: searchValue
    integer, intent(inout) :: jStart
    integer :: k, klo, khi

    !********************************************
    ! Check to see if x is less than the minimum
    ! value in the table
    !********************************************
    if (searchValue .lt. table(1)) then
        Locate2 = 1
        return
    end if

    !***************************************
    ! Check to see if x is greater than the
    ! maximum value in the table
    !***************************************
    if (searchValue .gt. table(size)) then
        Locate2 = size
        return
    end if

    klo = jStart
    khi = size

    do while((khi - klo) .gt. 1)
        k = (khi + klo) / 2
        if (table(k) .gt. searchValue) then
            khi = k
        else
            klo = k
        end if
    end do

    !*********************************
    ! klo and khi now bracket the
    ! searchvalue
    !*********************************
    Locate2 = klo
    jStart = klo
end function Locate2
!*****************************************************************
!
!****************************************************************
real(single) Function a151(n, x, xs)
    integer, intent(in) :: n
    real(single) :: x(:)
    
    real(single), pointer :: ws(:) => null()
    real(single) :: xm, xm0, xs, xc
    real(double) :: sum
    integer :: i
    real(single), parameter :: c = 1.5

    ws => allocateSingle(1, n, "ws")
    xm = medianValue(x, n)
    xs = smad(n, x, xm)
    xc = c * xs
    do
        xm0 = xm
        sum = 0.0_double
        do i = 1, n
            sum = sum + dble( min(xm0 + xc, max(xm0 - xc, x(i))))
        end do
        xm = sum / dble(n)
        if (abs(xm - xm0) .le. 1.0e-4) exit
    end do
    a151 = xm   

    deallocate(ws)
end function a151

!*****************************************************************
!
!****************************************************************
real(double) Function a152(n, x, xs)
    integer, intent(in) :: n
    real(double) :: x(:)
    
    real(single), pointer :: ws(:) => null()
    real(double) :: xm, xm0, xs, xc
    real(double) :: sum
    integer :: i
    real(double), parameter :: c = 1.5

    ws => allocateSingle(1, n, "ws")
    xm = medianValue(x, n)
    xs = smad(n, x, xm)
    xc = c * xs
    do
        xm0 = xm
        sum = 0.0_double
        do i = 1, n
            sum = sum + dble(min(xm0 + xc, max(xm0 - xc, x(i))))
        end do
        xm = sum / dble(n)
        if (abs(xm - xm0) .le. 1.0e-4) exit
    end do
    a152 = xm   

    deallocate(ws)
end function a152
!*****************************************************************
!
!****************************************************************
real(single) Function smad1(n, x, mu)
    integer, intent(in) :: n
    real(single) :: x(:)
    real(single), intent(in) :: mu
    real(single), pointer :: ws(:) => null()
    real(single) :: sm
    real(double) :: sum1
    integer :: i

    ws => allocateSingle(1, n, "ws")

    do i = 1, n
        ws(i) = abs(x(i) - mu)
    end do

    sm = medianValue(ws, n)
    if (sm .le. 0.0) then
        sum1 = sum(dble(ws(1 : n)))
        sm = sum1 / dble(n)
    end if
    
    smad1 = sm / 0.6745
    deallocate(ws)

end function smad1


!*****************************************************************
!
!****************************************************************
real(double) Function smad2(n, x, mu)
    integer, intent(in) :: n
    real(double) :: x(:)
    real(double), intent(in) :: mu
    real(double), pointer :: ws(:) => null()
    real(double) :: sm
    real(double) :: sum1
    integer :: i

    ws => allocateDouble(1, n, "ws")

    do i = 1, n
        ws(i) = abs(x(i) - mu)
    end do

    sm = medianValue(ws, n)

    if (sm .le. 0.0) then
        sum1 = sum(dble(ws(1 : n)))
        sm = sum1 / dble(n)
    end if
    
    smad2 = sm / 0.6745
    deallocate(ws)

end function smad2


!*****************************************************************
!
!****************************************************************
real(single) Function h151(n, x, xsc)
    integer, intent(in) :: n
    real(single) :: x(:)
    real(single), intent(out) :: xsc
    
    real(single), pointer :: ws(:) => null()
    real(single) :: a, c1, xm, xm0, xs, xs0, xc
    real(double) :: sum, sum2
    integer :: i
    real(single), parameter :: c = 1.5, beta = 0.778
    if (n < 1) then
        h151 = 0.0
        return
    end if
    ws => allocateSingle(1, n, "ws")
    c1 = c * sqrt(1.0 - 1.0 / n)

    xm = medianValue(x, n)
    xs = smad(n, x, xm)

    do
        xm0 = xm
        xs0 = xs
        sum = 0.0_double
        sum2 = 0.0_double
        xc = c1 * xs
        do i = 1, n
            a = min(xm + xc, max(xm - xc, x(i)))
            sum = sum + dble(a)
            sum2 = sum2 + (a - xm) * (a - xm)
        end do
        xm = sum / dble(n)
        xs = sqrt(sum2 / (beta * (n - 1)))
        if ( (abs(xm - xm0) .le. 1.0e-4 * xs0).or. &
              abs(xs / xs0 - 1.0) .le. 1.0e-4) exit
    end do
    h151 = xm
    xsc = xs        



    deallocate(ws)
end function h151



!*****************************************************************
!
!****************************************************************
real(double) Function h152(n, x, xsc)
    integer, intent(in) :: n
    real(double) :: x(:)
    real(double), intent(out) :: xsc
    
    real(double), pointer :: ws(:) => null()
    real(double) :: a, c1, xm, xm0, xs, xs0, xc
    real(double) :: sum, sum2
    integer :: i
    real(single), parameter :: c = 1.5, beta = 0.778
    if (n < 1) then
        h152 = 0.0_double
        return
    end if
    ws => allocateDouble(1, n, "ws")
    c1 = c * sqrt(1.0 - 1.0 / n)

    xm = medianValue(x, n)
    xs = smad(n, x, xm)

    do
        xm0 = xm
        xs0 = xs
        sum = 0.0_double
        sum2 = 0.0_double
        xc = c1 * xs
        do i = 1, n
            a = min(xm + xc, max(xm - xc, x(i)))
            sum = sum + dble(a)
            sum2 = sum2 + (a - xm) * (a - xm)
        end do
        xm = sum / dble(n)
        xs = sqrt(sum2 / (beta * (n - 1)))
        if ( (abs(xm - xm0) .le. 1.0e-4 * xs0) .or. &
              abs(xs / xs0 - 1.0) .le. 1.0e-4) exit
    end do
    h152 = xm
    xsc = xs        



    deallocate(ws)
end function h152





end module OmNumerics_mod









